from collections import defaultdict
from functools import reduce
from typing import Iterable, List, Any

from cbaxter1988_utils.log_utils import get_logger

logger = get_logger(__name__)


def map_list(func: callable, i: Iterable) -> List[Any]:
    """
    Runs a map function on iterable and returns a list

    :param func:
    :param i:
    :return:
    """
    return list(map(func, i))


def filter_list(func: callable, i: Iterable) -> List[Any]:
    """
    Runs a filter function on iterable and returns a list

    :param func:
    :param i:
    :return:
    """
    return list(filter(func, i))


def reduce_items(field, items: List[Any]) -> List[Any]:
    def reducer(acc, val):
        if isinstance(acc, dict):
            acc[val[field]].append(val)

        return acc

    return reduce(reducer, items, defaultdict(list))




from datetime import datetime


def get_utc_timestamp() -> float:
    return datetime.utcnow().timestamp()


def get_now():
    _now = datetime.now()
    return f'{_now.month}-{_now.day}-{_now.year} {_now.hour}:{_now.minute}::{_now.second}'



from cbaxter1988_utils.aws_dynamodb_utils import save_item, query_table_by_id, scan_table
from cbaxter1988_utils.daos.base_dao import BaseDAO


class DynamoDBDAO(BaseDAO):

    def __init__(self, table_id, **kwargs):
        self.table_id = table_id

    def save_item(self, item: dict):
        return save_item(table_id=self.table_id, item=item)

    def query_by_id(self, id: str):
        return query_table_by_id(table_id=self.table_id, id=id)

    def scan(self):
        return scan_table(table_id=self.table_id)



def assert_value_equals(value1, value2):
    assert value1 == value2


def assert_is_instance(testable, expected_instance):
    assert isinstance(testable, expected_instance)


def assert_is_subclass(testable, expected_class):
    assert issubclass(type(testable), expected_class)


def assert_len_eq(testable, expected_len):
    assert len(testable) == expected_len



import json
from dataclasses import dataclass, field, asdict

from cbaxter1988_utils.models import BaseModel

PROBLEM_DETAIL_HEADER_JSON = {
    "content-type": "application/problem+json"
}


@dataclass
class ProblemDetailModel(BaseModel):
    """


Based on Guidance published in RFC7807 - https://datatracker.ietf.org/doc/html/rfc7807


"type" (string) - A URI reference [RFC3986] that identifies the
problem type.  This specification encourages that, when
dereferenced, it provide human-readable documentation for the
problem type (e.g., using HTML [W3C.REC-html5-20141028]).  When
this member is not present, its value is assumed to be
"about:blank".

"title" (string) - A short, human-readable summary of the problem
type.  It SHOULD NOT change from occurrence to occurrence of the
problem, except for purposes of localization (e.g., using
proactive content negotiation; see [RFC7231], Section 3.4).

"status" (number) - The HTTP status code ([RFC7231], Section 6)
generated by the origin server for this occurrence of the problem.

"detail" (string) - A human-readable explanation specific to this
occurrence of the problem.

"instance" (string) - A URI reference that identifies the specific
occurrence of the problem.  It may or may not yield further
information if dereferenced.

The "status" member, if present, is only advisory; it conveys the
HTTP status code used for the convenience of the consumer.
Generators MUST use the same status code in the actual HTTP response,
to assure that generic HTTP software that does not understand this
format still behaves correctly.  See Section 5 for further caveats
regarding its use.

Consumers can use the status member to determine what the original
status code used by the generator was, in cases where it has been
changed (e.g., by an intermediary or cache), and when message bodies
persist without HTTP information.  Generic HTTP software will still
use the HTTP status code.


The "detail" member, if present, ought to focus on helping the client
correct the problem, rather than giving debugging information.

Consumers SHOULD NOT parse the "detail" member for information;
extensions are more suitable and less error-prone ways to obtain such
information.

Note that both "type" and "instance" accept relative URIs; this means
that they must be resolved relative to the document's base URI, as
per [RFC3986], Section 5.

    """
    type: str = field()
    type_base_uri: str = field(default='about:blank')
    title: str = field(default=None)
    status: int = field(default=None)
    detail: str = field(default=None)
    instance: str = field(default=None)

    header = PROBLEM_DETAIL_HEADER_JSON

    def __post_init__(self):
        self.type = f'{self.type_base_uri}/{self.type}'

    def to_json(self):
        return json.dumps(asdict(self))

    def to_xml(self):
        pass



import json
from typing import Any

from cbaxter1988_utils.log_utils import get_logger
from pika import BlockingConnection, URLParameters
from pika.adapters.blocking_connection import BlockingChannel
from pika.exceptions import ChannelClosed, ChannelClosedByBroker
from pika.exchange_type import ExchangeType

logger = get_logger(__name__)

PREFETCH_COUNT = 1


class PikaUtilsError(BaseException):
    """Utils Exception"""


class BlockingConnectionAdapter:

    def __init__(self, amqp_url):
        self.amqp_url = amqp_url
        self.url_params = URLParameters(url=self.amqp_url)
        self._prefetch_count = PREFETCH_COUNT

        self._prepare_connection()
        self._prepare_channel()

    def _prepare_connection(self):
        self._connection = get_blocking_connection(url=self.amqp_url)

    def _prepare_channel(self):
        if self._connection.is_closed:
            self._prepare_connection()
        else:
            self._channel = open_channel_from_connection(connection=self._connection)
            self._prepare_channel_qos()

    def get_channel(self):
        self._channel = open_channel_from_connection(connection=self.connection)
        return self._channel

    def _prepare_channel_qos(self):
        set_channel_qos(self._channel, prefetch_count=PREFETCH_COUNT)

    def _close_connection(self):
        logger.info(f"Closing Connection to '{self.amqp_url}'")
        return close_connection(self._connection)

    def _close_channel(self):
        return close_channel(channel=self._channel)

    def connect(self):
        if self._connection.is_closed:
            self._prepare_connection()
            self._prepare_channel()

    @property
    def channel(self):
        return self._channel

    @property
    def connection(self):
        return self._connection


class BasicPikaPublisher:
    logger = get_logger("BasicPublisher")

    def __init__(self, connection_adapter: BlockingConnectionAdapter, exchange, queue, routing_key):
        self.connection_adapter = connection_adapter
        self.exchange = exchange
        self.queue = queue
        self.routing_key = routing_key

    def publish_message(self, body: Any):
        try:
            bind_queue(
                connection=self.connection_adapter.connection,
                queue=self.queue,
                exchange=self.exchange,
                routing_key=self.routing_key
            )

            publish_message(
                connection=self.connection_adapter.connection,
                exchange=self.exchange,
                routing_key=self.routing_key,
                data=body
            )
        except ChannelClosedByBroker as err:
            logger.error(f'{err}')
            self.bind_routing_key(exchange=self.exchange, queue=self.queue, routing_key=self.routing_key)
            publish_message(
                self.connection_adapter.connection,
                exchange=self.exchange,
                routing_key=self.routing_key,
                data=body
            )

    def publish(self, routing_key, body):
        if self.connection_adapter.connection.is_open:
            self.connection_adapter.channel.basic_publish(exchange=self.exchange, routing_key=routing_key,
                                                          body=body)

    def declare_exchange(self, exchange=None):
        if exchange:
            self.exchange = exchange

        create_exchange(connection=self.connection_adapter.connection, exchange=self.exchange)

    def declare_queue(self, queue=None):
        if queue:
            self.queue = queue

        create_queue(connection=self.connection_adapter.connection, queue=self.queue)

    def bind_routing_key(self, exchange, queue, routing_key):
        self.declare_exchange(exchange)
        self.declare_queue(queue)
        if routing_key:
            self.routing_key = routing_key

        bind_queue(
            connection=self.connection_adapter.connection,
            queue=self.queue,
            exchange=self.exchange,
            routing_key=self.routing_key
        )

    @property
    def connection(self):
        return self.connection_adapter.connection

    @property
    def channel(self):
        return self.connection_adapter.channel


class PikaQueueConsumer:
    def __init__(self, amqp_url: str, queue_name: str, callback: callable):
        self.connection = get_blocking_connection(url=amqp_url)
        self._callback = callback
        self._queue_name = queue_name

    def consume(self, prefetch_count=None):
        channel = open_channel_from_connection(connection=self.connection)
        channel.basic_consume(queue=self._queue_name, on_message_callback=self._callback)
        if prefetch_count:
            set_channel_qos(channel=channel, prefetch_count=prefetch_count)

        try:
            channel.start_consuming()
        except Exception:
            logger.error(f"Exception Caught, Closing active channel: {channel.channel_number}")
            channel.stop_consuming()
            channel.close()
            raise


class BasicPikaConsumer:
    """
    Marked for deprecation, DO NOT USE, Use PikaQueueConsumer instead.
    
    """
    def __init__(
            self,
            connection_adapter: BlockingConnectionAdapter,
            queue,
            on_message_callback: callable
    ):

        self.connection_adapter = connection_adapter
        self.on_message_callback = on_message_callback
        self.queue = queue

        self.pika_queue_service_wrapper = PikaQueueServiceWrapper(amqp_url=self.connection_adapter.amqp_url)

    def _consume(self):
        if self.connection_adapter.connection.is_open and self.connection_adapter.channel.is_open:

            try:
                self.connection_adapter.channel.basic_consume(self.queue, self.on_message_callback)
                logger.info(f'Awaiting Message on channel: {self.connection_adapter.channel.channel_number}')
                self.connection_adapter.channel.start_consuming()


            except KeyboardInterrupt:
                self.connection_adapter.channel.stop_consuming()

            except ChannelClosedByBroker:
                raise
        else:
            self.connection_adapter.connect()

            self.connection_adapter.channel.basic_consume(self.queue, self.on_message_callback)
            self.connection_adapter.channel.start_consuming()

    def _validate_queue(self):
        logger.info(f"Validating Queue: '{self.queue}'")
        # self.pika_queue_service_wrapper.create_queue(
        #     queue=self.queue,
        #
        # )
        # conn = self.connection_adapter.connection
        # if not validate_queue(conn, queue=self.queue):
        #     logger.info(f"{self.queue} not present")
        #     create_queue(conn, queue=self.queue)

        #
        # else:
        #     logger.info(f'({self.queue}) has been validated')

    def run(self):
        # self._validate_queue()

        if self.connection_adapter.connection.is_closed:
            logger.info("Connection Closed, Reopening")
            self.connection_adapter.connect()

        if self.connection_adapter.channel.is_closed:
            logger.info("Channel Closed, Reopening")
            self.connection_adapter.get_channel()

        logger.info("Starting Consumer")
        self._consume()


class PikaQueueServiceWrapper:
    def __init__(self, amqp_url: str):
        self.amqp_url = amqp_url
        self.connection_adapter = BlockingConnectionAdapter(
            amqp_url=amqp_url
        )

        self._connection = get_blocking_connection(url=self.amqp_url)

    def create_queue(
            self,
            queue: str,
            dlq_queue: str = None,
            dlq_support: bool = False,
            dlq_exchange: str = None,
            dlq_routing_key: str = None
    ):
        queue_arguments = {}
        if self.connection_adapter.connection.is_closed:
            self.connection_adapter = BlockingConnectionAdapter(amqp_url=self.amqp_url)

        if dlq_support:
            create_queue(connection=self.connection, queue=dlq_queue)
            create_exchange(connection=self.connection, exchange=dlq_exchange)
            bind_queue(connection=self.connection, queue=dlq_queue, exchange=dlq_exchange,
                       routing_key=dlq_routing_key)

            queue_arguments['x-dead-letter-exchange'] = dlq_exchange
            queue_arguments['x-dead-letter-routing-key'] = dlq_routing_key

            logger.info(f'Prepared Pika Queue Arguments={queue_arguments}')

        create_queue(connection=self.connection, queue=queue, arguments=queue_arguments)

    def purge_queue(self, queue: str):
        purge_queue(connection=self.connection, queue=queue)

    def delete_queue(self, queue: str, if_empty: bool = True):
        delete_queue(connection=self.connection, queue=queue, if_empty=if_empty)

    # @property
    # def connection(self) -> BlockingConnection:
    #     if self.connection_adapter.connection.is_closed:
    #         return get_blocking_connection(url=self.connection_adapter.amqp_url)
    #     else:
    #         return self.connection_adapter.connection


def make_amqp_url(amqp_user, amqp_pw, amqp_host, amqp_port):
    return f"amqp://{amqp_user}:{amqp_pw}@{amqp_host}:{amqp_port}"


def validate_queue(connection: BlockingConnection, queue) -> bool:
    ch = open_channel_from_connection(connection)
    try:

        ch.queue_declare(queue=queue, passive=True)
        close_channel(ch)
        return True
    except ChannelClosed:
        close_channel(ch)

        return False


def create_queue(connection: BlockingConnection, queue, arguments: dict = None) -> bool:
    ch = open_channel_from_connection(connection)
    logger.info(f"Creating/Validating Queue='{queue}'")
    result = ch.queue_declare(queue=queue, arguments=arguments)

    close_channel(ch)
    return result


def delete_queue(connection: BlockingConnection, queue, if_empty=False, if_unused=False) -> bool:
    ch = open_channel_from_connection(connection)
    results = ch.queue_delete(queue=queue, if_empty=if_empty, if_unused=if_unused)
    close_channel(ch)
    if results:
        logger.debug(f"Removing Queue: {queue}")
        return True
    else:
        return False


def purge_queue(connection: BlockingConnection, queue):
    ch = open_channel_from_connection(connection)
    if ch.is_open:
        logger.debug(f"Purging Queue: {queue}")
        ch.queue_purge(queue=queue)
        return True

    else:
        raise PikaUtilsError(f"{ch.channel_number} is Closed")


def create_exchange(connection: BlockingConnection, exchange: str, exchange_type='direct') -> bool:
    ch = open_channel_from_connection(connection)
    if ExchangeType(exchange_type):
        try:
            logger.info(f'Creating/Validating Exchange={exchange}, ExchangeType={exchange_type}')
            ch.exchange_declare(exchange=exchange, exchange_type=ExchangeType(exchange_type).value, durable=True)

        except Exception:
            raise

        close_channel(ch)
        return True
    else:
        return False


def delete_exchange(connection: BlockingConnection, exchange, if_unused=False) -> bool:
    ch = open_channel_from_connection(connection)
    try:
        ch.exchange_delete(exchange=exchange, if_unused=if_unused)
    except Exception:
        raise

    close_channel(ch)
    return True


def bind_queue(connection: BlockingConnection, queue, exchange, routing_key) -> bool:
    ch = open_channel_from_connection(connection)
    try:
        logger.info(f"Binding Exchange='{exchange}' to Queue='{queue}' with Key='{routing_key}')")
        ch.queue_bind(queue=queue, exchange=exchange, routing_key=routing_key)

    except Exception:
        raise

    close_channel(ch)
    return True


def unbind_queue(connection: BlockingConnection, queue, exchange, routing_key) -> bool:
    ch = open_channel_from_connection(connection)
    try:
        ch.queue_unbind(queue=queue, exchange=exchange, routing_key=routing_key)
    except Exception:
        raise

    close_channel(ch)
    return True


def open_channel_from_connection(connection: BlockingConnection) -> BlockingChannel:
    if connection.is_open:
        channel = connection.channel()
        logger.debug(f"Opened Channel: {channel.channel_number}")
        return channel
    else:

        raise PikaUtilsError(f"{connection} is Closed")


def close_channel(channel: BlockingChannel):
    if channel.is_open:
        logger.debug(f"Closing Channel: {channel.channel_number}")
        channel.close()
        return True
    else:
        return False


def close_connection(connection: BlockingConnection):
    if connection.is_open:
        logger.info(f"Closing Connection: {str(connection)}")
        connection.close()
        return True
    else:
        return False


def set_channel_qos(channel: BlockingChannel, prefetch_count):
    if channel.is_open:
        channel.basic_qos(prefetch_count=prefetch_count)
        return True
    else:
        raise PikaUtilsError(f"{channel.channel_number} is Closed")


def get_blocking_connection(url) -> BlockingConnection:
    logger.info(f"Connecting to URL: '{url}'")
    return BlockingConnection(
        parameters=URLParameters(url=url)
    )


def publish_message(connection: BlockingConnection, exchange: str, routing_key: str, data: Any) -> bool:
    ch = open_channel_from_connection(connection)
    if ch.is_open:
        try:
            logger.info(f"Publishing Message: {data}")
            if isinstance(data, bytes):
                ch.basic_publish(exchange=exchange, routing_key=routing_key, body=data)

            if isinstance(data, dict):
                ch.basic_publish(exchange=exchange, routing_key=routing_key, body=json.dumps(data).encode())

            close_channel(ch)
            return True
        except Exception:
            raise
    else:

        raise PikaUtilsError(f"{ch.channel_number} is Closed")


def acknowledge_message(channel: BlockingChannel, delivery_tag):
    channel.basic_ack(delivery_tag=delivery_tag)


def nacknowledge_message(channel: BlockingChannel, delivery_tag):
    channel.basic_nack(delivery_tag=delivery_tag)


def make_basic_pika_publisher(amqp_url, exchange, queue, routing_key) -> BasicPikaPublisher:
    adapter = BlockingConnectionAdapter(amqp_url=amqp_url)
    return BasicPikaPublisher(connection_adapter=adapter, queue=queue, exchange=exchange, routing_key=routing_key)


def make_basic_pika_consumer(amqp_url, queue, on_message_callback: callable) -> BasicPikaConsumer:
    adapter = BlockingConnectionAdapter(amqp_url=amqp_url)
    return BasicPikaConsumer(connection_adapter=adapter, queue=queue, on_message_callback=on_message_callback)


def make_pika_queue_consumer(amqp_url, queue, on_message_callback: callable) -> PikaQueueConsumer:
    return PikaQueueConsumer(amqp_url=amqp_url, queue_name=queue, callback=on_message_callback)



import os


def get_env_strict(key):
    """
    Gets the eviornment variable based on the key, Raise KeyError in env var is invalid.
    :param key: enviornment variable
    :return:
    """
    try:
        return os.environ[key]
    except KeyError:
        print(f"Environment Variable: '{key}' Not Set ")
        if raise_exception:
            raise KeyError(f"Environment Variable: '{key}' Not Set")


def get_env(key, default_value=None):
    """
    Gets the eviornment variable based on the key, returns default value if key is not found

    :param key:
    :param default_value:
    :return:
    """
    return os.getenv(key, default_value)

def set_env(key, val):
    """
    Sets the env var to the val.

    :param key:
    :param val:
    :return:
    """
    os.putenv(key, val)


import requests


def get_html_page(url) -> str:
    return requests.get(url).text


def get_request(url):
    return requests.get(url)


def post_request(url, body):
    response = requests.post(url=url, json=body, headers={"Content-Type": "application/json"})
    return response



from unittest.mock import patch

from cbaxter1988_utils.aws_cognito_utils import add_new_user_basic, list_user_pools, delete_user
from pytest import fixture


@fixture
def aws_client_mock():
    with patch("cbaxter1988_utils.aws_cognito_utils.get_cognito_idp_client") as client_mock:
        yield client_mock


@fixture
def test_user_name():
    return 'test_user'


@fixture
def test_user_pool_id():
    return 'us-east-1_fRkg83NZI'


@fixture
def test_user_email():
    return 'cbaxtertech@gmail.com'


def test_add_new_user(
        aws_client_mock,
        test_user_name,
        test_user_email,
        test_user_pool_id
):
    add_new_user_basic(
        pool_id=test_user_pool_id,
        user_email=test_user_email,
        username=test_user_name
    )

    aws_client_mock.assert_called()


def test_delete_user(
        aws_client_mock,
        test_user_pool_id,
        test_user_name
):
    delete_user(pool_id=test_user_pool_id, username=test_user_name)
    aws_client_mock.assert_called()


def test_list_user_pools(aws_client_mock):
    list_user_pools()

    aws_client_mock.assert_called()



from cbaxter1988_utils.src.daos.dynamodb_dao import DynamoDBDAO
from cbaxter1988_utils.src.daos.mongodb_dao import MongoDBDAO

__all__ = [
    DynamoDBDAO,
    MongoDBDAO,
]

class Factory:

    def get_mongodb_dao(self):
        return MongoDBDAO(
            host="",
            port=0,
            db="",
            collection=""
        )


import base64
import json

from cbaxter1988_utils.aws_utils import get_event_bridge_client


def _get_client():
    return get_event_bridge_client()


def publish_event(
        source: str,
        event_bus_name: str,
        event_bus_arn: str,
        event_type: str,
        event_context: dict,
        encode: bool = False
):
    client = _get_client()
    return client.put_events(
        Entries=[
            {
                "Source": source,
                "EventBusName": event_bus_name,
                "Resources": [
                    event_bus_arn
                ],
                "DetailType": event_type,
                "Detail": event_context if not encode else base64.b64encode(f'{json.dumps(event_context)}'.encode()),
            }

        ]
    )


class EventManager:
    def __int__(self):
        pass



import json

import pytest
from flask import Response
from cbaxter1988_utils.flask_utils import build_json_response, build_problem_detail_response


@pytest.fixture()
def json_response_dict():
    return build_json_response(
        status=200,
        payload={"msg": "test_message"}
    )


@pytest.fixture()
def json_response_str():
    return build_json_response(
        status=200,
        payload=json.dumps({"msg": "test_message"})
    )


@pytest.fixture()
def problem_detail_response():
    return build_problem_detail_response(
        instance=ValueError("Invalid Value"),
        detail="Account Overdrafted",
        title="AccountOverDraft",
        type_name="account_overdraft",
        http_status_code=200,
        exstensions=[
            {
                "account_id": "test_id"
            }
        ]
    )


def _test_build_json_response_1(json_response_dict):
    test_subject: Response = json_response_dict
    assert isinstance(test_subject, Response)
    assert test_subject.status == "200 OK"
    assert test_subject.content_type == "application/json"
    assert test_subject.response == [b'{"msg": "test_message"}']


def _test_build_json_response_2(json_response_str):
    test_subject: Response = json_response_str
    assert isinstance(test_subject, Response)
    assert test_subject.status == "200 OK"
    assert test_subject.content_type == "application/json"
    assert test_subject.response == [b'{"msg": "test_message"}']


def _test_build_problem_detail_response(problem_detail_response):
    test_subject: Response = problem_detail_response

    assert isinstance(test_subject, Response)
    assert test_subject.status == "200 OK"
    assert test_subject.content_type == "application/problem+json"
    assert test_subject.response == [b'{"type": "about:blank/account_overdraft", "type_base_uri": "about:blank", "title": "AccountOverDraft", "status": 200, "detail": "Account Overdrafted", "instance": "Invalid Value"}']
    assert test_subject.json['type'] == "about:blank/account_overdraft"
    assert test_subject.json['status'] == 200


from dataclasses import dataclass

import pytest
from tests.const import VAR_NAME


@pytest.fixture
def person_class():
    """
    Sample Person dataclass that can used in tests.

    :return:
    """

    @dataclass
    class Person:
        f_name: str

    return Person


@pytest.fixture
def person(person_class):
    """
    Sample Person instance that can be used in tests.

    :param person_class:
    :return:
    """
    return person_class(
        f_name=VAR_NAME
    )


@pytest.fixture()
def number_list():
    """
    List of 10 numbers

    :return:
    """
    return [
        i + 1 for i in range(10)
    ]



from typing import List

from boto3.dynamodb.conditions import Key
from cbaxter1988_utils.aws_utils import get_dynamo_table


def _get_table(table_id):
    return get_dynamo_table(table_id)


def save_item(table_id: str, item: dict) -> dict:
    table = _get_table(table_id=table_id)
    return table.put_item(Item=item)


def query_table_by_id(table_id: str, id: str) -> List[dict]:
    table = _get_table(table_id=table_id)
    response = table.query(KeyConditionExpression=Key('id').eq(id))
    return response['Items']


def scan_table(table_id: str):
    table = _get_table(table_id=table_id)
    response = table.scan()
    return response['Items']



from cbaxter1988_utils.log_utils import get_logger

logger = get_logger(__name__)


def log_invocation(func):
    def inner_func(*args, **kwargs):
        logger.info(f'Logging Function Execution: "{func.__name__}" Called With: (kwargs={kwargs})')
        return func(*args, **kwargs)

    return inner_func



import logging

formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
stream_handler.setLevel(logging.INFO)

global_logger = logging.getLogger("GlobalLogger")
global_logger.setLevel(logging.INFO)
global_logger.addHandler(stream_handler)


def get_logger(name=__name__, level=logging.INFO):
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)

    _logger = logging.getLogger(name)
    _logger.setLevel(level)
    _logger.addHandler(stream_handler)
    return _logger


def get_global_logger():
    return global_logger



import json
from base64 import b64encode, b64decode
from dataclasses import is_dataclass, asdict
from datetime import datetime
from enum import Enum
from typing import Any, Union
from uuid import UUID


class CustomJSONEncoder(json.JSONEncoder):

    def default(self, o: Any) -> Any:
        if is_dataclass(o):
            return asdict(o)

        if isinstance(o, Enum):
            return o.value

        if isinstance(o, UUID):
            return str(o)

        if isinstance(o, bytes):
            return o.decode()

        if isinstance(o, datetime):
            return o.isoformat()

        if isinstance(o, dict):
            return self.default(o)

        return json.JSONEncoder.default(self, o)


def serialize_object(o: Any, b64_encode: bool = False) -> Union[str, bytes]:
    serialized_object = json.dumps(o, cls=CustomJSONEncoder)
    if b64_encode:
        return b64encode(serialized_object.encode())
    else:
        return serialized_object


def decode_b64_object(b64_data: bytes) -> dict:
    return json.loads(b64decode(b64_data).decode())


def rebuild_object(object_factory_func: callable, data: dict) -> Any:
    return object_factory_func(**data)


def rebuild_decode_b64_object(object_factory_func: callable, b64_data: bytes) -> Any:
    data = decode_b64_object(b64_data=b64_data)

    return rebuild_object(object_factory_func=object_factory_func, data=data)



import random
import string
from dataclasses import dataclass


@dataclass
class PasswordPolicy:
    pass


@dataclass
class BasicPasswordPolicy:
    length: int = 1


@dataclass
class AdvancedPasswordPolicy:
    length: int = 1
    alphabets_count: int = 1
    digits_count: int = 1
    special_characters_count: int = 1
    upper_case_count: int = 1
    lower_case_count: int = 1


def make_basic_password_policy(length: int) -> BasicPasswordPolicy:
    return BasicPasswordPolicy(
        length=length
    )


def make_advanced_password_policy(
        length: int,
        alphabets_count: int,
        digits_count: int,
        special_characters_count: int,
        lower_case_count: int,
        upper_case_count: int

) -> AdvancedPasswordPolicy:
    return AdvancedPasswordPolicy(
        length=length,
        alphabets_count=alphabets_count,
        digits_count=digits_count,
        special_characters_count=special_characters_count,
        lower_case_count=lower_case_count,
        upper_case_count=upper_case_count,
    )


def generate_random_password_basic(password_policy: BasicPasswordPolicy):
    characters = list(string.ascii_letters + string.digits + "!@#$%^&*()")

    random.shuffle(characters)

    password = []
    for i in range(password_policy.length):
        password.append(random.choice(characters))

    random.shuffle(password)

    return "".join(password)


def generate_random_password_advanced(password_policy: AdvancedPasswordPolicy):
    alphabets = list(string.ascii_letters)
    digits = list(string.digits)
    special_characters = list("!@#$%^&*()")
    characters = list(string.ascii_letters + string.digits + "!@#$%^&*()")

    ## length of password from the user
    length = password_policy.length

    ## number of character types
    alphabets_count = password_policy.alphabets_count
    digits_count = password_policy.digits_count
    special_characters_count = password_policy.special_characters_count

    characters_count = alphabets_count + digits_count + special_characters_count

    ## check the total length with characters sum count
    ## print not valid if the sum is greater than length
    if characters_count > length:
        print("Characters total count is greater than the password length")
        return

    ## initializing the password
    password = []

    ## picking random alphabets
    for _ in range(alphabets_count):
        password.append(random.choice(alphabets))

    ## picking random digits
    for _ in range(digits_count):
        password.append(random.choice(digits))

    ## picking random alphabets
    for _ in range(special_characters_count):
        password.append(random.choice(special_characters))

    for _ in range(password_policy.upper_case_count):
        password.append(random.choice(string.ascii_uppercase))

    for _ in range(password_policy.lower_case_count):
        password.append(random.choice(string.ascii_lowercase))

    ## if the total characters count is less than the password length
    ## add random characters to make it equal to the length
    if characters_count < length:
        random.shuffle(characters)
        for _ in range(length - characters_count):
            password.append(random.choice(characters))

    ## shuffling the resultant password
    random.shuffle(password)

    ## converting the list to string
    ## printing the list
    return "".join(password)



from cbaxter1988_utils.daos.base_dao import DAOMixinInterface, BaseDAO
from cbaxter1988_utils.pagination_utils import (
    BasePage,
    BasePaginator
)
from cbaxter1988_utils.pymongo_utils import (
    get_client,
    get_collection,
    get_item,
    update_item,
    add_item,
    get_database,
    query_items,
    delete_item,
    DEFAULT_ITEM_KEY,
    InsertOneResult,
    DeleteResult,
    UpdateResult,
    Cursor,
    get_page_from_collection,
    get_pages_from_collection,

)


class _MongoDBPymongoMixin(DAOMixinInterface):
    def query_items(self, query):
        return query_items(collection=self.collection, query=query)

    def get_item(self, item_id, item_key=DEFAULT_ITEM_KEY) -> Cursor:
        return get_item(collection=self.collection, item_id=item_id, item_key=item_key)

    def update_item(self, item_id, new_values, item_key=DEFAULT_ITEM_KEY) -> UpdateResult:
        return update_item(collection=self.collection, item_id=item_id, new_values=new_values, item_key=item_key)

    def add_item(self, item: dict, key_id='_id') -> InsertOneResult:
        return add_item(collection=self.collection, item=item, key_id=key_id)

    def delete_item(self, item_id, item_key=DEFAULT_ITEM_KEY) -> DeleteResult:
        return delete_item(collection=self.collection, item_id=item_id, item_key=item_key)

    def scan_items(self):
        return query_items(collection=self.collection, query={})

    def get_page(self, query, limit_per_page=500, last_item_id=None) -> BasePage:
        return get_page_from_collection(self.collection, query=query, limit=limit_per_page, last_item_id=last_item_id)

    def get_pages(self, query, limit_per_page=500) -> BasePaginator:
        return get_pages_from_collection(self.collection, query=query, page_size=limit_per_page)


class MongoDBDAO(BaseDAO, _MongoDBPymongoMixin):
    def __init__(self, host: str, port: int, db: str, collection: str):
        self._host = host
        self._port = port

        self._client = get_client(db_host=self._host, db_port=self._port)

        self.db = db
        self.collection = collection

    @property
    def collection(self):
        return self._collection

    @collection.setter
    def collection(self, collection: str):
        self._collection = get_collection(self._db, collection)

    @property
    def db(self):
        return self._db

    @db.setter
    def db(self, db):
        self._db = get_database(client=self._client, db_name=db)

    @property
    def client(self):
        return self._client



from cbaxter1988_utils.auth_utils import (
    generate_random_password_advanced,
    make_advanced_password_policy
)
from cbaxter1988_utils.aws_utils import get_cognito_idp_client
from cbaxter1988_utils.log_utils import get_logger

logger = get_logger(__name__)


def add_new_user_basic(
        pool_id: str,
        username: str,
        user_email: str,
        temp_password_len: int = 8,
        temp_password_alphabets_count: int = 1,
        temp_password_digits_count: int = 1,
        temp_password_special_characters_count: int = 1,
        temp_password_upper_case_count: int = 1,
        temp_password_lower_case_count: int = 1,
):
    """
    This function Creates a new pool user for the specified UserPool.

    This function only uses the 'email' UserAttributes when creating the user.



    :param pool_id: ID of the AWS cognito Pool
    :param username: Desired username
    :param user_email: User email address
    :param temp_password_len:
    :param temp_password_alphabets_count:
    :param temp_password_digits_count:
    :param temp_password_special_characters_count:
    :param temp_password_upper_case_count:
    :param temp_password_lower_case_count:
    :return:
    """
    client = get_cognito_idp_client()

    temp_password = generate_random_password_advanced(
        make_advanced_password_policy(
            length=temp_password_len,
            alphabets_count=temp_password_alphabets_count,
            digits_count=temp_password_digits_count,
            special_characters_count=temp_password_special_characters_count,
            upper_case_count=temp_password_upper_case_count,
            lower_case_count=temp_password_lower_case_count
        )
    )
    logger.info(f"Created Temp Password='{temp_password}' for '{user_email}'")
    return client.admin_create_user(
        UserPoolId=pool_id,
        Username=username,
        TemporaryPassword=temp_password,
        UserAttributes=[
            {
                'Name': 'email',
                'Value': user_email
            },
        ]
    )


def delete_user(pool_id: str, username: str):
    client = get_cognito_idp_client()
    result = client.admin_delete_user(
        UserPoolId=pool_id,
        Username=username
    )

    logger.info(f"Removed User='{username}' from UserPool={pool_id}")
    return result


def list_user_pools():
    client = get_cognito_idp_client()
    return client.list_user_pools(
        MaxResults=50
    )



from abc import ABC, abstractmethod


class DAOMixinInterface(ABC):

    @abstractmethod
    def query_items(self, query):
        """Abstract Method"""

    @abstractmethod
    def get_item(self, item_id, item_key):
        """Abstract Method"""

    @abstractmethod
    def update_item(self, item_id, new_values, item_key):
        """Abstract Method"""

    @abstractmethod
    def add_item(self, item: dict):
        """Abstract Method"""

    @abstractmethod
    def delete_item(self, item_id, item_key):
        """Abstract Method"""

    @abstractmethod
    def scan_items(self):
        """Abstract Method"""


class BaseDAO:
    pass



import http
import json
from http import HTTPStatus
from typing import Union, List

import flask
from cbaxter1988_utils.models.http_models.problem_detail_rfc_7807_model import ProblemDetailModel, PROBLEM_DETAIL_HEADER_JSON


def build_problem_detail_response(
        http_status_code: Union[HTTPStatus, int],
        detail: str,
        type_name: str,
        title: str,
        instance: BaseException,
        exstensions: List[dict]
) -> flask.Response:
    """
    Builds a ProblemDetail response based on RFC 7807

    :param http_status_code: The HTTP status code ([RFC7231]
    :param detail: A human-readable explanation specific to this
occurrence of the problem.
    :param type_name: A URI reference [RFC3986] that identifies the problem type.
    :param title:  A short, human-readable summary of the problem type.
    :param instance:  A URI reference that identifies the specific
occurrence of the problem.
    :return:
    """
    model = ProblemDetailModel(
        type=type_name,
        status=http_status_code,
        detail=detail,
        instance=str(instance),
        title=title,
    )
    for exstension in exstensions:
        for key, val in exstension.items():
            setattr(model, key, val)

    return flask.Response(
        status=http_status_code,
        headers=PROBLEM_DETAIL_HEADER_JSON,
        response=model.to_json()

    )


def build_json_response(status: http.HTTPStatus, payload: Union[dict, str]) -> flask.Response:
    """
    Returns Flask.Response as content-type JSON

    :param status:
    :param payload:
    :return:
    """
    if isinstance(payload, dict):
        return flask.Response(
            status=status,
            headers={"content-type": "application/json"},
            response=json.dumps(payload)
        )

    if isinstance(payload, str):
        return flask.Response(
            status=status,
            headers={"content-type": "application/json"},
            response=payload
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, List

"""
A Collection of tools that can be used for handling Pagination. 

"""


class IBasePage(ABC):
    """
    Interface for BasePage
    """

    @abstractmethod
    def add_item(self, item: Any):
        pass


class IPaginator(ABC):
    """
    Interface for Pagniator

    """

    @abstractmethod
    def add_page(self, item: IBasePage):
        pass

    @abstractmethod
    def get_page(self, page_number: int):
        pass


@dataclass()
class BasePage(IBasePage):
    """
    Basic Page
    """
    page_id: int
    items: List[Any]
    next_page: int
    previous_page: int
    item_count: int

    def __post_init__(self):
        self.next_page = self.page_id + 1
        self.previous_page = self.page_id - 1
        if self.previous_page == 0:
            self.previous_page = 1

    @staticmethod
    def make_pages(items: List[Any], page_size: int, sort: bool = True, reverse: bool = True) -> List['BasePage']:
        try:
            if sort:
                items.sort()
        except TypeError:
            pass

        if reverse:
            items.reverse()

        page_count = round(len(items) / page_size)

        pages = []
        for i in range(page_count):
            i += 1
            page_items = []
            for idx in range(len(items)):

                if len(page_items) < page_size:
                    page_items.append(items.pop())

            page = BasePage(
                page_id=i,
                items=page_items,
                next_page=0,
                previous_page=0,
                item_count=page_size
            )
            pages.append(page)

        return pages

    def add_item(self, item: Any):
        pass

    def get_page(self, page_number: int):
        pass


class BasePaginator(IPaginator):
    """
    Basic Paginator that can be used to create pages
    """

    def __init__(self, **kwargs):
        self._page_map = {}
        self._page_list = []

        self._active_page = None
        self._total_pages = None
        self.total_records = 0

    def _check_page_map_list(self):
        if len(self._page_list) == len(self._page_map.keys()):
            return True
        else:
            return False

    def add_page(self, page: BasePage):
        """
        Adds new page to paginator

        :param page:
        :return:
        """
        if self._check_page_map_list():
            self._page_map[page.page_id] = page
            self._page_list.append(page)
            return True
        else:
            print(f"Unable to add page {page}")
            return False

    def remove_page(self, page_id: int):
        """
        Removes page from paginator

        :param page_id:
        :return:
        """
        page = self._page_map[page_id]
        self._page_list.remove(page)
        del self._page_map[page_id]

    def add_pages(self, pages: List[BasePage]):
        _ = [
            self.add_page(page=page)
            for page in pages
        ]

    def get_page(self, page_number: int) -> BasePage:
        return self._page_map[page_number]

    def make_pages(self, items: List[Any], page_size: int, reverse: bool = True, sort: bool = True):
        pages = BasePage.make_pages(
            items=items,
            page_size=page_size,
            reverse=reverse,
            sort=sort,
        )
        self.add_pages(pages=pages)
        return self

    @property
    def pages(self):
        return self._page_list

    @property
    def page_count(self):
        if len(self._page_map.keys()) == len(self.pages):
            return len(self.pages)

    def paginate(self):
        cur_page = 0
        total_pages = 10
        while cur_page <= total_pages:
            pass



import boto3
from cbaxter1988_utils.environment_utils import get_env

"""
Collection of Utilities interacting with the boto3 library

"""


def get_boto3_client(service):
    """
    Gets new boto3 client

    :param service: AWS service id, example: s3
    :return:
    """
    return boto3.client(
        service,
        region_name=get_env(key="AWS_DEFAULT_REGION"),
        aws_access_key_id=get_env(key="AWS_ACCESS_KEY_ID"),
        aws_secret_access_key=get_env(key="AWS_SECRET_ACCESS_KEY")
    )


def get_event_bridge_client():
    """
    Returns EventBridge client.

    :return:
    """
    return get_boto3_client('events')


def get_lambda_client():
    """
    Returns EventBridge client.

    :return:
    """
    return get_boto3_client('lambda')


def get_s3_client():
    """
    Returns S3 client.

    :return:
    """
    return get_boto3_client(service="s3")


def get_dynamo_client():
    """
    Returns DynamoDB client.

    :return:
    """
    return get_boto3_client(service="dynamo")


def get_cloudwatch_client():
    """
    Returns Cloudwatch client.

    :return:
    """
    return get_boto3_client(service='cloudwatch')


def get_dynamo_table(table_id: str):
    """
    Returns DynamoDB table.

    :return:
    """
    resource = boto3.resource('dynamodb')

    return resource.Table(table_id)


def get_cognito_idp_client():
    return get_boto3_client('cognito-idp')



import json


def extract_body_from_event(event):
    request_body = None
    if event.get("body"):
        if isinstance(event['body'], str):
            request_body = json.loads(event['body'])
        else:
            request_body = event['body']
    else:
        if isinstance(event, str):
            request_body = json.loads(event)
        else:
            request_body = event

    return request_body


def extract_params_from_event(event):
    if event.get("pathParameters"):
        return event['pathParameters']



from cbaxter1988_utils.core_utils import clone_object, check_list_len


def test_clone_object(person):
    new_object = clone_object(person)
    assert id(person) != id(new_object)


def test_check_list_len(number_list):
    result = check_list_len(items=number_list, expected_len=10)
    assert result is True



def create_dimension(name, value):
    return {
        'Name': name,
        'Value': value
    }


def update_metric(client, name_space, metric_name, dimensions, value=None, values=None):
    if value:
        return client.put_metric_data(
            Namespace=name_space,
            MetricData=[
                {
                    'MetricName': metric_name,
                    'Dimensions': dimensions,
                    'Value': int(value)

                },
            ]
        )



import pytest
from dataclasses import dataclass
from cbaxter1988_utils.dataclass_utils import model_to_dict



def test_model_to_dict(person):
    # invocation
    person_dict = model_to_dict(person)

    # validation
    assert isinstance(person_dict, dict)



import os
from unittest.mock import patch

from cbaxter1988_utils.pika_utils import (
    make_basic_pika_publisher,
    PikaQueueConsumer,
    PikaQueueServiceWrapper
)
from pytest import fixture

AMQP_USER = os.getenv("AMQP_USER", 'guest')
AMQP_PW = os.getenv("AMQP_PW", 'guest')
AMQP_HOST = os.getenv("AMQP_HOST", '192.168.1.5')
AMQP_PORT = os.getenv("AMQP_PORT", 5672)

AMQP_URL = os.getenv("AMPQ_URL", f"amqp://{AMQP_USER}:{AMQP_PW}@{AMQP_HOST}:{AMQP_PORT}")


@fixture
def test_queue_name():
    return 'PIKA_DEV_TEST_QUEUE'


@fixture
def test_queue_exchange_name():
    return 'PIKA_DEV_EXCHANGE'


@fixture
def test_queue_routing_key():
    return 'PIKA_DEV_ROUTING_KEY'


@fixture
def test_dlq_name():
    return 'PIKA_DEV_DLQ_TEST_QUEUE'


@fixture
def test_dlq_routing_key_name():
    return 'PIKA_DEV_DLQ_ROUTING_KEY'


@fixture
def test_dlq_exchange_name():
    return 'PIKA_DEV_DLQ_EXCHANGE'


@fixture
def blocking_connection_adapter_mock():
    with patch("cbaxter1988_utils.pika_utils.BlockingConnectionAdapter", autospec=True) as connector_mock:
        yield connector_mock


@fixture
def basic_pika_publisher_testable(test_queue_name, test_queue_exchange_name, test_queue_routing_key):
    return make_basic_pika_publisher(
        amqp_url=AMQP_URL,
        queue=test_queue_name,
        routing_key=test_queue_routing_key,
        exchange=test_queue_exchange_name
    )


@fixture
def pika_queue_service_wrapper_testable():
    return PikaQueueServiceWrapper(
        amqp_url=AMQP_URL,

    )


def test_basic_publisher_when_publishing_message(blocking_connection_adapter_mock, basic_pika_publisher_testable):
    publisher = basic_pika_publisher_testable
    publisher.publish_message(body={
        "test": "test"
    })
    blocking_connection_adapter_mock.assert_called()


def test_pika_queue_service_wrapper_when_creating_queue(
        blocking_connection_adapter_mock,
        pika_queue_service_wrapper_testable,
        test_queue_name,
        test_dlq_name,
        test_dlq_routing_key_name,
        test_dlq_exchange_name
):
    queue_builder = pika_queue_service_wrapper_testable

    queue_builder.create_queue(
        queue=test_queue_name,
        dlq_support=True,
        dlq_queue=test_dlq_name,
        dlq_routing_key=test_dlq_routing_key_name,
        dlq_exchange=test_dlq_exchange_name
    )
    blocking_connection_adapter_mock.assert_called()


def test_pika_queue_service_wrapper_when_deleting_queue(
        blocking_connection_adapter_mock,
        pika_queue_service_wrapper_testable,
        test_queue_name
):
    queue_wrapper = pika_queue_service_wrapper_testable
    queue_wrapper.delete_queue(
        queue=test_queue_name,
        if_empty=False
    )

    blocking_connection_adapter_mock.assert_called()


def test_pika_queue_service_wrapper_when_purging_queue(
        blocking_connection_adapter_mock,
        pika_queue_service_wrapper_testable,
        test_queue_name
):
    queue_wrapper = pika_queue_service_wrapper_testable
    queue_wrapper.purge_queue(queue=test_queue_name)
    blocking_connection_adapter_mock.assert_called()


def test_pika_queue_consumer():
    def on_message_callback(ch, method, properties, body):
        print(ch, method, properties, body)

    consumer = PikaQueueConsumer(amqp_url=AMQP_URL, queue_name='TEST_QUEUE', callback=on_message_callback)

    consumer.consume()


from copy import deepcopy

from typing import List, Any

"""
Collection of random  utilities 
"""


def check_list_len(items: List[Any], expected_len: int) -> bool:
    """
    Checks the length of the given list

    :param items: List of any type of objects.
    :param expected_len: the expected length of the list.
    :return:
    """
    if len(items) == expected_len:
        return True
    else:
        return False


def clone_object(item: Any) -> Any:
    """
    Creates a copy of the given item.

    :param item: Any object expected to be cloned
    :return:
    """
    return deepcopy(item)



import pytest
from cbaxter1988_utils.pagination_utils import BasePaginator, BasePage


@pytest.fixture
def items():
    return [
        {"test": "test"},
        {"test": "test1"},
        {"test": "test2"},
        {"test": "test3"},
        {"test": "test4"},
        {"test": "test5"},
        {"test": "test6"},
        {"test": "test7"},
        {"test": "test8"},
        {"test": "test9"},
        {"test": "test10"},
    ]


def test_base_paginator(number_list):
    """
    Tests the creation of pages with items

    :param number_list:
    :return:
    """
    # Set Up
    page_items_expectation = [4, 5, 6]
    page_count_expectation = 3

    paginator = BasePaginator()

    # invocation
    paginator.make_pages(items=number_list, page_size=3)
    page = paginator.get_page(page_number=2)

    # Validation
    assert isinstance(page, BasePage)
    assert paginator.page_count == page_count_expectation
    assert page.items == page_items_expectation


def test_base_pagiation(items):
    # Set Up
    paginator = BasePaginator()
    page_items_expectation = [{'test': 'test'}, {'test': 'test1'}]
    page_count_expectation = 6

    # Invocation
    paginator.make_pages(items=items, page_size=2)
    page = paginator.get_page(page_number=1)

    # Validation
    assert isinstance(page, BasePage)
    assert page.items == page_items_expectation
    assert paginator.page_count == page_count_expectation



from dataclasses import dataclass, asdict
from uuid import uuid5, UUID, NAMESPACE_DNS

from cbaxter1988_utils.pymongo_utils import (
    add_item,
    get_client,
    get_collection,
    get_database,
    get_item,
    safe_update_item,
    safe_delete_item,
    get_mongo_client_w_auth,
    add_admin_database_super_user,
    remove_admin_database_user
)
from pytest import fixture


@fixture()
def mongo_client_w_auth():
    return get_mongo_client_w_auth(
        db_host="192.168.1.5",
        db_port=27017,
        db_username='admin',
        db_password='pimpin12'
    )


@fixture()
def internal_database():
    client = get_mongo_client_w_auth(
        db_host="192.168.1.5",
        db_port=27017,
        db_username='admin',
        db_password='pimpin12'
    )
    db = get_database(client=client, db_name='LOCAL_BOE_MVP')

    return db


@dataclass
class PersonModel:
    _id: UUID
    first_name: str
    last_name: str
    version: int = 1


def _test_occ_update():
    person_model = PersonModel(
        _id=uuid5(NAMESPACE_DNS, f'Elijah'),
        first_name='Elijah',
        last_name='Baxter'
    )

    client = get_client(db_host='192.168.1.5', db_port=27017)
    database = get_database(client=client, db_name='test_db')
    collection = get_collection(database=database, collection='test_collection')
    item_id = UUID('be3a2142-c794-5817-9023-4764c36d0cdd')
    elijah_id = UUID('9607cc96-c17e-5af4-91cb-52aaeb182177')

    try:
        add_item(collection=collection, item=asdict(person_model))
    except Exception as err:
        print(err)

    item = get_item(collection=collection, item_id=item_id)
    data = item.next()

    expected_version = data.get("version")
    resp = safe_update_item(
        collection=collection,
        item_id=item_id,
        expected_version=expected_version,
        new_values={"last_name": "Baxter", "age": 6}
    )

    resp = safe_delete_item(collection=collection, item_id=elijah_id, expected_version=1)


# def _test_add_database_admin_user(internal_database):
#     add_database_admin_user(
#         database=internal_database,
#         username='admin',
#         password='admin'
#     )


def _test_add_item(internal_database):
    collection = get_collection(database=internal_database, collection="LOCAL_BANK_ACCOUNT_AGGREGATE_TABLE")
    result = add_item(collection=collection, item={"test": "test"})
    print(result.acknowledged)


def _test_add_admin_database_super_user(mongo_client_w_auth):
    add_admin_database_super_user(client=mongo_client_w_auth, username='cbaxter', password='password')


def _test_remove_admin_database_user(mongo_client_w_auth):
    remove_admin_database_user(client=mongo_client_w_auth, username='cbaxter')



import base64
import datetime
import json
from dataclasses import dataclass
from enum import Enum
from uuid import UUID

from cbaxter1988_utils.serialization_utils import (
    serialize_object,
    rebuild_object,
    decode_b64_object,
    rebuild_decode_b64_object
)
from pytest import fixture


class CarTypeEnum(Enum):
    sedan = 1
    suv = 2


@dataclass
class Car:
    make: str
    model: str
    car_type: CarTypeEnum
    id: UUID


@dataclass
class Robot:
    data: bytes
    b64_data: bytes
    created: datetime.datetime = datetime.datetime.now()


def rebuild_car(make: str, model: str, car_type: int, id: str):
    return Car(
        make=make,
        model=model,
        car_type=CarTypeEnum(car_type),
        id=UUID(id)
    )


@fixture
def test_car_object():
    return Car(
        car_type=CarTypeEnum.sedan,
        make='bmw',
        model='5-series',
        id=UUID("c6a3d4c0-9805-4f03-aa20-4c511f7c75e1")
    )


@fixture
def test_robot_object():
    return Robot(
        data=b'TEST_DATA',
        b64_data=base64.b64encode(b'TEST_DATA'),
        created=datetime.datetime.fromisoformat("2021-12-22T23:23:56.763252")
    )


@fixture
def test_robot_dict():
    return {
        'data': 'TEST_DATA',
        'b64_data': 'VEVTVF9EQVRB',
        'created': '2021-12-22T23:23:56.763252'
    }


@fixture
def test_car_object_dict():
    return {
        "make": "bmw",
        "model": "5-series",
        "car_type": 1,
        "id": "c6a3d4c0-9805-4f03-aa20-4c511f7c75e1"
    }


@fixture
def b64_test_object():
    return b'eyJtYWtlIjogImJtdyIsICJtb2RlbCI6ICI1LXNlcmllcyIsICJjYXJfdHlwZSI6IDEsICJpZCI6ICJjNmEzZDRjMC05ODA1LTRmMDMtYWEyMC00YzUxMWY3Yzc1ZTEifQ=='


def test_serialize_object_when_decoding_base64(test_car_object, b64_test_object):
    data_bytes = serialize_object(test_car_object, b64_encode=True)
    assert data_bytes == b64_test_object
    assert isinstance(data_bytes, bytes)


def test_decode_b64_object(b64_test_object):
    data = decode_b64_object(b64_data=b64_test_object)
    assert isinstance(data, dict)
    assert data == {'make': 'bmw', 'model': '5-series', 'car_type': 1, 'id': 'c6a3d4c0-9805-4f03-aa20-4c511f7c75e1'}


def test_rebuild_object(test_car_object):
    json_string = serialize_object(test_car_object)

    data_dict = json.loads(json_string)

    car = rebuild_object(object_factory_func=rebuild_car, data=data_dict)
    assert isinstance(car, Car)


def test_rebuild_decode_b64_object(b64_test_object, test_car_object):
    car = rebuild_decode_b64_object(b64_data=b64_test_object, object_factory_func=rebuild_car)
    assert isinstance(car, Car)
    assert car == test_car_object


def test_serialize_object_when_processing_car_object(test_car_object, test_car_object_dict):
    json_string = serialize_object(test_car_object)
    assert json.loads(json_string) == test_car_object_dict


def test_serialize_object_when_processing_robot_object(test_robot_object, test_robot_dict):
    json_string = serialize_object(test_robot_object)
    print(json_string)
    json_data = json.loads(json_string)
    assert json_data == test_robot_dict

    #



from typing import Union, List
from uuid import UUID

from bson import ObjectId
from cbaxter1988_utils.log_utils import get_logger
from cbaxter1988_utils.pagination_utils import BasePaginator, BasePage
from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.database import Database
from pymongo.errors import DuplicateKeyError
from pymongo.results import DeleteResult, UpdateResult, InsertOneResult, InsertManyResult

DEFAULT_ITEM_KEY = "_id"

logger = get_logger(__name__)


class InvalidUpdate(BaseException):
    """Exception for InvalidUpdate"""


class InvalidDelete(BaseException):
    """Exception for InvalidUpdate"""


def make_objectid() -> ObjectId:
    return ObjectId()


def get_client(db_host, db_port=27017) -> MongoClient:
    return MongoClient(host=db_host, port=db_port)


def get_mongo_client_w_auth(
        db_host: str,
        db_username: str,
        db_password: str,
        db_port: int = 27017,
) -> MongoClient:
    return MongoClient(host=db_host, port=db_port, username=db_username, password=db_password)


def get_database(client: MongoClient, db_name: str) -> Database:
    return client[db_name]


def get_collection(database: Database, collection: str) -> Collection:
    return database[collection]


def query_items(collection: Collection, query: dict) -> Cursor:
    return collection.find(query)


def scan_items(collection: Collection) -> Cursor:
    return collection.find()


def update_item(collection: Collection, item_id: Union[str, int], new_values: dict,
                item_key=DEFAULT_ITEM_KEY) -> UpdateResult:
    return collection.update_one(
        {
            f"{item_key}": item_id
        },
        {
            "$set": new_values
        }
    )


def add_item(collection: Collection, item: dict, key_id='_id') -> InsertOneResult:
    try:
        if not key_id == '_id':
            item['_id'] = item[key_id]

        return collection.insert_one(document=item)

    except DuplicateKeyError:
        raise


def delete_item(collection: Collection, item_id: Union[str, int], item_key=DEFAULT_ITEM_KEY) -> DeleteResult:
    return collection.delete_one(
        {
            f"{item_key}": item_id
        },

    )


def get_item(collection: Collection, item_id: Union[str, int, UUID], item_key=DEFAULT_ITEM_KEY) -> Cursor:
    return collection.find(
        {
            f"{item_key}": item_id
        }
    )


def get_page_from_collection(collection: Collection, query: dict, limit, last_item_id=None) -> BasePage:
    total_records = collection.find(query).count()

    if last_item_id:
        query.update({"_id": {"$gt": last_item_id}})

    cursor = collection.find(query).limit(limit)

    last_item_id = _get_last_id_from_cursor(cursor=cursor)

    return NewPage(
        cursor=cursor,
        last_id=last_item_id,
        total_items=total_records
    )


def get_pages_from_collection(collection: Collection, query: dict, page_size: int) -> BasePaginator:
    paginator = BasePaginator()
    items = collection.find(query)
    paginator.make_pages(items=items, page_size=page_size)
    return paginator


def _get_last_id_from_cursor(cursor):
    _cursor = clone_item(cursor)
    cursor_list = list(_cursor)
    try:
        return cursor_list[len(cursor_list) - 1].get("_id")

    except IndexError:
        return None


def _get_cursor_count(cursor):
    _cursor = clone_item(cursor)
    cursor_list = list(_cursor)
    return len(cursor_list)


def add_many_items(collection: Collection, items: List[dict], ordered: bool = True) -> InsertManyResult:
    return collection.insert_many(
        documents=items,
        ordered=ordered
    )


def check_for_items(collection: Collection):
    collection.aggregate()


def safe_update_item(collection: Collection, item_id: Union[UUID, str], expected_version: int, new_values: dict):
    """
    Preforms Optimistic Check by validating item version. If the expected_version does
    not match the document version, the transation will not be applied.

    Useful in distributed systems requiring safe updates to mongo records


    :param collection:
    :param item_id:
    :param expected_version:
    :param new_values:
    :return:
    """
    result = collection.update_one(
        filter={"version": expected_version, "_id": item_id},
        update={
            "$set": new_values,
            "$inc": {
                "version": 1
            }
        }
    )

    if result.raw_result.get("updatedExisting") is False:
        item = collection.find(filter={"_id": item_id}).next()
        raise InvalidUpdate(
            f"Version Number Mismatch, Expected:  '{expected_version}', Document Version: '{item.get('version')}'"
        )

    return result


def safe_delete_item(collection: Collection, item_id: Union[UUID, str], expected_version: int) -> DeleteResult:
    result = collection.delete_one(
        filter={"version": expected_version, "_id": item_id},

    )
    if result.raw_result.get("n") == 0:
        item = collection.find(filter={"_id": item_id}).next()
        raise InvalidDelete(
            f"Version Number Mismatch, Expected:  '{expected_version}', Document Version: '{item.get('version')}'"
        )

    return result


def add_database_user(database: Database, username, password, roles: List[str] = None):
    if roles is None:
        roles = []

    return database.command("createUser", username, pwd=password, roles=roles)


def add_database_user_rw(database: Database, username, password):
    return database.command("createUser", username, pwd=password, roles=['readWrite'])


def add_database_user_db_owner(database: Database, username, password):
    """
    The database owner can perform any administrative action on the database. This role combines the privileges granted by the readWrite, dbAdmin and userAdmin roles.


    :param database:
    :param username:
    :param password:
    :return:
    """
    return database.command("createUser", username, pwd=password, roles=['dbOwner'])


def add_admin_database_super_user(client: MongoClient, username, password):
    admin_db = get_database(db_name='admin', client=client)
    return admin_db.command(
        "createUser", username, pwd=password, roles=[
            "userAdminAnyDatabase", "readWriteAnyDatabase"])


def authenticate_database_basic(database: Database, username: str, password: str):
    return database.authenticate(username, password=password)


def remove_admin_database_user(client: MongoClient, username):
    admin_db = get_database(db_name='admin', client=client)
    return admin_db.command('dropUser', username)



from pathlib import Path

from setuptools import find_packages
from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='cbaxter1988_utils',
    version='0.7.0',
    url='https://github.com/cbaxter1988/utils.git',
    description='A Package containing my utils',
    author='Courtney S Baxter Jr',
    author_email='cbaxtertech@gmail.com',
    packages=find_packages(),
    install_requires=required,
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True
)



import json

from cbaxter1988_utils.aws_utils import get_boto3_client


def get_secrets_manager_client():
    return get_boto3_client('secretsmanager')


def get_credentials(secret_name: str) -> dict:
    client = get_secrets_manager_client()
    get_secret_value_response = client.get_secret_value(
        SecretId=secret_name
    )

    if 'SecretString' in get_secret_value_response:
        secret = get_secret_value_response['SecretString']

        return json.loads(secret)



from cbaxter1988_utils.auth_utils import (
    generate_random_password_basic,
    generate_random_password_advanced,
    make_advanced_password_policy,
    make_basic_password_policy
)

from pytest import fixture


@fixture
def basic_password_policy():
    return make_basic_password_policy(
        length=8
    )


@fixture
def advanced_password_policy():
    return make_advanced_password_policy(
        length=8,
        digits_count=1,
        upper_case_count=1,
        lower_case_count=1,
        special_characters_count=2,
        alphabets_count=1
    )


def test_generate_random_password_basic(
        basic_password_policy
):
    pw = generate_random_password_basic(
        password_policy=basic_password_policy
    )
    assert len(pw) == 8


def test_generate_random_password_advanced(
        advanced_password_policy
):
    pw = generate_random_password_advanced(
        password_policy=advanced_password_policy

    )

    assert len(pw) >= 8



from dataclasses import asdict, is_dataclass

"""
Collection of python dataclass utilites 
"""


def model_to_dict(data_class) -> dict:
    """
    Converts the given Dataclass to a dictionary

    Raises TypeError if the given type is not a python dataclass

    :param data_class:
    :return:
    """
    if is_dataclass(data_class):
        return asdict(data_class)
    else:
        raise TypeError("Invalid Type Provided")
